﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using HeuristicLab.Core;
using HeuristicLab.MainForm;
using HeuristicLab.MainForm.WindowsForms;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Encodings.ParameterConfigurationEncoding.Views {
  public partial class CreateExperimentDialogV2 : Form {
    private readonly Progress progress;
    private readonly IAlgorithm algorithm;
    private ParameterConfigurationTree parameterConfigurationTree;
    private bool createBatchRun;
    private int repetitions;
    private IEngine engine;
    private Task generation;
    private CancellationTokenSource cts;

    public Experiment Experiment { get; private set; }

    public CreateExperimentDialogV2() : this(null, null) { }
    public CreateExperimentDialogV2(IAlgorithm algorithm, IEnumerable<IEngine> engines) {
      InitializeComponent();
      if (algorithm != null && engines != null) {
        this.algorithm = (IAlgorithm)algorithm.Clone();
        foreach (var engine in engines) engineComboBox.Items.Add(engine);
        if (engineComboBox.SelectedItem == null) {
          engineComboBox.SelectedIndex = 0;
          engine = (IEngine)engineComboBox.SelectedItem;
        }
      }
      createBatchRun = createBatchRunCheckBox.Checked;
      repetitions = (int)repetitionsNumericUpDown.Value;
      progress = new Progress {
        CanBeCanceled = true,
        ProgressState = ProgressState.Finished
      };
      progress.CancelRequested += (sender, args) => cts.Cancel();
      progress.ProgressValueChanged += (sender, args) => progress.Status = string.Format("Generating experiment. Please be patient. ({0} %)", (int)(progress.ProgressValue * 100));
      cts = new CancellationTokenSource();
    }

    private void SetEnabledStateOfControls(bool state) {
      createBatchRunCheckBox.Enabled = state;
      repetitionsNumericUpDown.Enabled = state;
      engineComboBox.Enabled = state;
      okButton.Enabled = state;
      cancelButton.Enabled = state;
    }

    private void StartProgressView() {
      if (InvokeRequired) {
        Invoke(new Action(StartProgressView));
      } else {
        SetEnabledStateOfControls(false);
        progress.ProgressValue = 0;
        progress.ProgressState = ProgressState.Started;
      }
    }

    private void FinishProgressView() {
      if (InvokeRequired) {
        Invoke(new Action(FinishProgressView));
      } else {
        progress.Finish();
        SetEnabledStateOfControls(true);
      }
    }

    private void ReportError(Exception e) {
      if (InvokeRequired) {
        Invoke(new Action<Exception>(ReportError), e);
      } else {
        ErrorHandling.ShowErrorDialog(this, "An error occurred generating the experiment.", e);
      }
    }

    private void CreateExperimentDialogV2_Load(object sender, System.EventArgs e) {
      if (algorithm.Problem != null) {
        parameterConfigurationTreeView.Content = parameterConfigurationTree = new ParameterConfigurationTree(algorithm, algorithm.Problem);
        parameterConfigurationTree.ExperimentGenerationProgressChanged += (o, args) => progress.ProgressValue = parameterConfigurationTree.ExperimentGenerationProgress;
        new ProgressView(parameterConfigurationTreeView, progress) { CancelTimeoutMs = 10000 };
      }
    }

    private void CreateExperimentDialogV2_FormClosing(object sender, FormClosingEventArgs e) {
      if (generation != null) e.Cancel = true;
    }

    private void createBatchRunCheckBox_CheckedChanged(object sender, System.EventArgs e) {
      repetitionsNumericUpDown.Enabled = createBatchRun = createBatchRunCheckBox.Checked;
    }

    private void repetitionsNumericUpDown_Validated(object sender, System.EventArgs e) {
      if (repetitionsNumericUpDown.Text == string.Empty)
        repetitionsNumericUpDown.Text = repetitionsNumericUpDown.Value.ToString();
      repetitions = (int)repetitionsNumericUpDown.Value;
    }

    private void engineComboBox_SelectedIndexChanged(object sender, System.EventArgs e) {
      engine = (IEngine)engineComboBox.SelectedItem;
    }

    private void calculateCombinationsButton_Click(object sender, EventArgs e) {
      combinationsTextBox.Text = parameterConfigurationTree.GetCombinationCount(0).ToString();
    }

    private void okButton_Click(object sender, System.EventArgs e) {
      generation = Task.Factory.StartNew(() => {
        StartProgressView();
        var engineAlgorithm = algorithm as EngineAlgorithm;
        if (engineAlgorithm != null) engineAlgorithm.Engine = engine;
        try {
          Experiment = createBatchRun ? parameterConfigurationTree.GenerateExperiment(engineAlgorithm, true, repetitions, cts.Token)
                                      : parameterConfigurationTree.GenerateExperiment(engineAlgorithm, false, 0, cts.Token);
        }
        finally { FinishProgressView(); }
      }, cts.Token);
      generation.ContinueWith(t => {
        generation = null;
        this.DialogResult = DialogResult.OK;
        this.Close();
      }, TaskContinuationOptions.OnlyOnRanToCompletion);
      generation.ContinueWith(t => {
        generation = null;
        if (t.IsCanceled) cts = new CancellationTokenSource();
        if (t.IsFaulted) {
          ReportError(t.Exception.Flatten());
          t.Exception.Flatten().Handle(ex => true);
        }
      }, TaskContinuationOptions.NotOnRanToCompletion);
    }
  }
}
